/* vc_ub.c

   Vertex cover upper bounds

   Ambros Marzetta

   94-11-28
   For more information see our WWW page at
   
         http://wwwjn.inf.ethz.ch/ambros/zram.html
	 
 * $Source$
 * $Revision$
 * $Date$
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.1  1998/04/15  08:47:29  marzetta
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include "z_bb.h"
#include "vc_base.h"
#include "vc_ub.h"

#define TRACEUB

static vc_NodeP vc_UpperBoundRest(vc_GlobalP global, vc_NodeP np,
				  MINT *head, MINT *tail)
{
  /* Modifies np and returns it. */
  MINT pt, other;
  
  pt = np->vc_numPt - 1;
  while (pt >= 0) {
    if (vc_active(np, pt)) { /* the point is active */
      other = SomeAdjacent(global, np, pt);
      assert(vc_active(np, other));
      vc_Mark(global, np, pt, &head);
      if (vc_active(np, other)) 
	vc_Mark(global, np, other, &head);
      MarkHeadTail(global, np, &head, &tail);
    }
    pt --;
  } /* while */
  assert(np->vc_numAct == 0);

  UnmarkPoints(global, np);
  
  np->vc_lb = np->vc_numMarked;
  return np;
} /* vc_UpperBoundRest */

static MINT *degree1list;
static long degree1listsize;

vc_NodeP vc_UpperBound1(vc_GlobalP global, vc_NodeP np)
{
  /* Allocate a (not minimal) vertex cover as upper bound for the solution. */
  /* Approximation with 2 points per edge */
  /* Process degree 1 points first. */
  /* This upper bound is at most twice the solution. */
  MINT *head, *tail;

  np = (vc_NodeP) z_NewCopyNode((z_NodeP) np);

  if (np->vc_numEdge == 0)
    return np;	/* This is a solution node */

  z_ReSizePointer((void *)&degree1list,
		&degree1listsize, np->vc_numPt * sizeof(MINT));
  head = tail = degree1list;

  return vc_UpperBoundRest(global, np, head, tail);
} /* vc_UpperBound1 */

static int ub2depth;

vc_NodeP vc_UpperBound2(vc_GlobalP global, vc_NodeP np)
{
  /* Find a minimal vertex cover for a half size subgraph.
     Compute a (not minimal) vertex cover as upper bound for the solution.
     */
  vc_NodeP npr, sol, sol1, sol2;
  int qsize, i;
  MINT *head, *tail;

  if (ub2depth > 0 || np->vc_numAct < 100)
    return vc_UpperBound1(global, np);
  
  sol1 = vc_UpperBound1(global, np);
  
  npr = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
  Restrict(global, npr, 1);	/* restrict to left subgraph */
  if (npr->vc_numAct == np->vc_numAct) {
				/* np is already a left subgraph from lb6 */
#ifdef TRACEUB
    puts("UB: left");
#endif
    (void) vc_FreeNode(npr);
    return sol1;
  }
  if (npr->vc_numAct == 0) {
				/* np is already a right subgraph from lb6 */
#ifdef TRACEUB
    puts("UB: right");
#endif
    (void) vc_FreeNode(npr);
    return sol1;
  }
  qsize = 10000000 / np->len;
  ub2depth ++;
  sol = (vc_NodeP) z_BranchBoundMixedFirst(global->vc_procsetnum,
					     (z_bb_GlobalDataP) global,
					     (z_NodeP)npr, NULL, qsize, 0);

  z_ReSizePointer((void *)&degree1list,
		&degree1listsize, np->vc_numPt * sizeof(MINT));
  head = tail = degree1list;

  np = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
				/* do not modify original np */
  for (i = sol->vc_numPt - 1; i >= 0; i--)
    if (vc_degree(sol)[i] == MARK_DEGREE
	&& vc_degree(np)[i] != MARK_DEGREE)
      vc_Mark(global, np, i, &head);
  /*printf("UB: pt=%ld, act%ld, lact=%ld, lmark=%ld, ",
	 (long) np->vc_numPt, (long) np->vc_numAct,
	 (long) npr->vc_numAct, (long) sol->vc_numMarked);*/
  (void) vc_FreeNode(npr);
  (void) vc_FreeNode(sol);

  sol2 = vc_UpperBoundRest(global, np, head, tail);
#ifdef TRACEUB
  printf("UB: %ld %ld\n", (long) sol1->vc_lb, (long) sol2->vc_lb);
#endif
  if (sol1->vc_lb < sol2->vc_lb) { /* upper bound 1 is better */
    (void) vc_FreeNode(sol2);
    sol2 = sol1;
  } else
    (void) vc_FreeNode(sol1);
  ub2depth --;
  return sol2;
} /* vc_UpperBound2 */

